% Script to score the *test* results of the benchmark
% and compute statistically significant differences
% in the ranked list.

% Isabelle Guyon -- August 2003 -- isabelle@clopinet.com

% DISCLAIMER: ALL INFORMATION, SOFTWARE, DOCUMENTATION, AND DATA ARE PROVIDED "AS-IS" 
% THE ORGANIZERS DISCLAIMS ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
% THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. 
% IN NO EVENT SHALL ISABELLE GUYON AND/OR OTHER BENCHMARK ORGANIZERS BE LIABLE FOR ANY SPECIAL, 
% INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER ARISING OUT OF OR IN CONNECTION 
% WITH THE USE OR PERFORMANCE OF BENCHMARK SOFTWARE, DOCUMENTS, MATERIALS, PUBLICATIONS, OR 
% INFORMATION MADE AVAILABLE. 

where_my_data_is='.';           % This is where the 5 directories with the datasets are
resudir=[where_my_data_is '\Results\']; % Where the result files are stored (.resu, .conf, .feat)
truthdir=[where_my_data_is '\Truth\']; % Where the truth files are stored (.labels)
scoredir=[where_my_data_is '\Score\']; % Where the scoring results are stored.
warning off MATLAB:MKDIR:DirectoryExists;
mkdir(scoredir);
fout=fopen([scoredir '\Compar-' date '.score'],'w') ;

% Identify the method, dataset, and result types:
[method, dataset] = identify(resudir);
Nd=length(dataset);
Nm=length(method);

% Loop over datasets and methods:
for i=1:Nd
    fprintf(fout, '\n** %s **\n', upper(dataset{i}));
    
    % Read the target values:
    Y=load([truthdir dataset{i} '_test.labels']); % Binary classification label
    patnum=length(Y);
    F=load([truthdir dataset{i} '_feat.labels']); % Binary label +1 for good feature, -1 for probe.
    featnum=length(F);
    
    % Read the test results:
    Y_resu=cell(Nm, 1); % Predicted value of the classifier (binary).
    Y_score=cell(Nm, 1); % Discriminant coefficient.
    F_used=cell(Nm, 1); % Indices of the selected features.
    for j=1:Nm
        fn=[resudir method{j} '\' dataset{i} '_test.resu'];
        Y_resu{j}=get_resu(fn);
        fn=[resudir method{j} '\' dataset{i} '_test.conf'];
        Y_conf=get_resu(fn);
        Y_score{j}=Y_conf.*Y_resu{j};
        fn=[resudir method{j} '\' dataset{i} '.feat'];
        F_used{j}=get_resu(fn);
        if isempty(F_used{j})
           F_used{j}=1:featnum; % Assume all the features have been used
        end
    end
    
    % Compute simple statistics:
    errate=-1*ones(Nm,1); % Balanced error rate.
    aucval=-1*ones(Nm,1); % Area under the curve
    frac_feat=-1*ones(Nm,1); % Fraction of features used among the features given.
    frac_probe=-1*ones(Nm,1); % Fraction of probes found among the features used.
    fprintf(fout, '\nUnsorted results:\nmethod\terrate\tauc\tfrac_feat\tfrac_probe\n');
        for j=1:Nm
        % Balanced error rate:
        if ~isempty(Y_resu{j})
            errate(j)=balanced_errate(Y_resu{j}, Y);
        end
        % AUC:
        if ~isempty(Y_score{j})
            aucval(j) = auc(Y_score{j}, Y);
        end
        % Fraction of features used
        used_num=length(F_used{j});
        frac_feat(j)=used_num/featnum;
        % Fraction of probes in the features used
        probe_num=length(find(F(F_used{j})==-1));
        if probe_num==0, used_num=1; end
        frac_probe(j)=probe_num/used_num;
        
        if errate(j)~=-1
            fprintf(fout, '%s\t%5.2f\t%5.2f\t%5.4f\t%5.4f\n',method{j},100*errate(j),100*aucval(j),100*frac_feat(j),100*frac_probe(j));
        end
    end
               
    % Restrict the method set to those having effectively results:
    eff_idx=find(errate~=-1);
    eff_method=method(eff_idx);
    Nem=length(eff_idx);
    
    % Fill in the count matrices (weighted number of errors on diag., 
    % weighted number of errors one makes and the other does not off diag.)
    R=zeros(Nem, Nem);
    aidx=find(Y>0);
    bidx=find(Y<0);
    na=length(aidx);
    nb=length(bidx);
    alpha=patnum/(2*na); % weight for class positive class errors
    beta=patnum/(2*nb); % weight for negative class errors
    for jj=1:Nem
        j=eff_idx(jj);
        yj=Y_resu{j};
        for kk=1:Nem
            k=eff_idx(kk);
            yk=Y_resu{k};
            if jj==kk
                Na=sum(yj(aidx)<0);
                Nb=sum(yj(bidx)>0);
                R(jj,jj)=alpha*Na+beta*Nb; % total number of errors
            else
                Ra=sum(yj(aidx)<0 & yk(aidx)>0);
                Rb=sum(yj(bidx)>0 & yk(bidx)<0);
                R(jj,kk)=alpha*Ra+beta*Rb;
            end
        end
    end
    fprintf(fout, '\nError counts:\n');
    write_matrix(fout, R, eff_method);
    
    % Compute the balanced McNemar statistic:
    M=zeros(Nem, Nem);
    for jj=1:Nem
        for kk=1:Nem
            den=sqrt(R(jj,kk)+R(kk,jj));
            if den==0
                z=0;
            else
                z=(R(kk,jj)-R(jj,kk))/den;
            end
            M(jj,kk)=0.5*(1-erf(z/sqrt(2)));
        end
        M(jj,jj)=R(jj,jj)/patnum; % balanced error rate on the diagonal
    end
    fprintf(fout, '\nMcNemar statistics (errate on diagonal):\n');
    write_matrix(fout, M, eff_method);
    
    % Fill a matrix with test results:
    T=zeros(Nem, Nem);
    alpha=0.05;
    T(M < alpha)=1;
    T(M >(1-alpha))=-1;
    for jj=1:Nem, T(jj,jj)=0; end % Zeros on the diagonal
    fprintf(fout, '\nMcNemar test results:\n');
    write_matrix(fout, T, eff_method);
    
    % Substitute the zeros in the test matrix using the fraction of
    % features used:
    theta=0.05;
    for jj=1:Nem
        j=eff_idx(jj);
        for kk=1:Nem
            k=eff_idx(kk);
            if jj~=kk & T(jj,kk)==0,
                rel_diff=2*abs(frac_feat(k)-frac_feat(j))/(frac_feat(k)+frac_feat(j));
                if rel_diff>theta,
                    if(frac_feat(j)<frac_feat(k))
                        T(jj,kk)=1;
                    elseif(frac_feat(j)>frac_feat(k))
                        T(jj,kk)=-1;
                    else
                        fprintf(fout, '%f  %f %f\n',rel_diff,frac_feat(j),frac_feat(k));
                    end % otherwise it stays 0.
                end
            end
        end
    end
    fprintf(fout, '\nUsing the fraction of features used:\n');
    write_matrix(fout, T, eff_method);
    
    % Substiture the zeros in the test matrix using the fraction of probes
    % in the features used:
    for jj=1:Nem
        j=eff_idx(jj);
        for kk=1:Nem
            k=eff_idx(kk);
            if jj~=kk & T(jj,kk)==0,
                if(frac_probe(j)<frac_probe(k))
                        T(jj,kk)=1;
                elseif(frac_probe(j)>frac_probe(k))
                        T(jj,kk)=-1;
                end % otherwise it stays 0.
            end
        end
    end
    fprintf(fout, '\nUsing the fraction of probes:\n');
    write_matrix(fout, T, eff_method);
    
    % Rank never-the-less according to score if everything failed:
    for jj=1:Nem
        j=eff_idx(jj);
        for kk=1:Nem
            k=eff_idx(kk);
            if jj~=kk & T(jj,kk)==0,
                if(M(jj,jj)<M(kk,kk))
                        T(jj,kk)=1;
                elseif(M(jj,jj)>M(kk,kk))
                        T(jj,kk)=-1;
                end % otherwise it stays 0.
            end
        end
    end
    fprintf(fout, '\nUsing the error rate:\n');
    write_matrix(fout, T, eff_method);
    
    % Rank using the average line score (number of times it outperforms
    % another classifier.
    
    C=sum(T);
    for jj=1:Nem, T(jj,jj)=C(jj); end
    [cs, is]=sort(C);
    
    Smethod=eff_method(is);
    ST=T(is,is);
    fprintf(fout, '\nSame matrix sorted (scores on the diagonal):\n');
    write_matrix(fout, ST, Smethod);
    
    fprintf(fout, '\nSorted results:\nmethod\terrate\tauc\tfrac_feat\tfrac_probe\n');
    for jj=1:Nem
        j=eff_idx(is(jj));
        fprintf(fout, '%s\t%5.2f\t%5.2f\t%5.4f\t%5.4f\n',method{j},100*errate(j),100*aucval(j),100*frac_feat(j),100*frac_probe(j));
    end    
end % The end!

fclose(fout);